#include <stdlib.h>
#include <math.h>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <string>
#include <vector>
#include "Sequence.h"
#include "BinaryPatternSequence.h"

using namespace std;

//*********************************************************************************************************//
BinaryPatternSequence::BinaryPatternSequence() {
	has_data = 1;
	starting_positions = NULL;
	binary_patsequence = NULL;
	pat_matrix = new PatternMatrix;	
};
//*********************************************************************************************************//
BinaryPatternSequence::~BinaryPatternSequence() {
	has_data = 0;
	if ((starting_positions != NULL) && (binary_patsequence != NULL)) { FreeMemory (); }
	delete pat_matrix;
};
//*********************************************************************************************************//
void BinaryPatternSequence::FreeMemory () {
	has_data = 0;
	delete [ ] starting_positions;
	delete [ ] binary_patsequence;

	starting_positions = NULL;
	binary_patsequence = NULL;
};
//*********************************************************************************************************//
void BinaryPatternSequence::SetSize (unsigned long sequence_length) {
	has_data = 1;
	array_size = sequence_length;

	binary_patsequence = new bool[array_size];
	starting_positions = new int[array_size];

	for (unsigned int i = 0; i < sequence_length; ++i) {
		binary_patsequence[i]   = 0;
		starting_positions[i]   = 0;
	}
};
//*********************************************************************************************************//
void BinaryPatternSequence::SetArray( unsigned long counter, std::string plus_pattern, std::string combined_sequence ) {
	unsigned long match1(0);
	unsigned long match2(0);

	char buffer[50];	
	string temp = "";
	string mins_pattern = "";
	string match_temp_pattern = "";

	Sequence pattern;
	plus_pattern = pattern.UpperCase(plus_pattern);
	mins_pattern = pattern.RevComp(plus_pattern);	
	pattern_length = pattern.PromoterLength(plus_pattern);
	original_pattern = plus_pattern;

	for (unsigned int i = 0; i < combined_sequence.size() - pattern_length; ++i) {
		match_temp_pattern = combined_sequence.substr(i, pattern_length);
		match1 = 0;
		match2 = 0;
		if (match_temp_pattern.find("$", 0) == string::npos) {
			match1 = pattern.MatchIUPACPattern(match_temp_pattern, plus_pattern);
			match2 = pattern.MatchIUPACPattern(match_temp_pattern, mins_pattern);
		}

		if (match1 > 0 || match2 > 0) {
			if (match1 > 0) { starting_positions[i] =  pattern_length; } 
			if (match2 > 0) { starting_positions[i] = (-1 * pattern_length); }
			for (unsigned int j = 0; j < pattern_length; ++j) {
				binary_patsequence[i + j] = 1;
			}
		} 
	}

	sprintf(buffer, "%lu", counter);
	temp = buffer;
	original_pattern = plus_pattern + "_nro" + temp;
	(*pat_matrix).SetPatternMatrix(pattern_length);
};
//*********************************************************************************************************//
double BinaryPatternSequence::CalculateDistance(BinaryPatternSequence * compare, unsigned long sequence_number) {
	double true_pos(0);		// 1 = 1
	double false_pos(0);		// 0 = 1
	double true_neg(0);		// 0 = 0
	double false_neg(0);		// 1 = 0
	double total_size(0);

	double distance(0);
	double nominator(0);
	double divider(0);

	for (unsigned int i = 0; i < array_size ; ++i) {
		if ((binary_patsequence[i] == 0) && ((*compare).binary_patsequence[i] == 0)) { true_neg = true_neg + 1;  }	
		if ((binary_patsequence[i] == 1) && ((*compare).binary_patsequence[i] == 0)) { false_neg = false_neg + 1; }
		if ((binary_patsequence[i] == 0) && ((*compare).binary_patsequence[i] == 1)) { false_pos = false_pos + 1; }
		if ((binary_patsequence[i] == 1) && ((*compare).binary_patsequence[i] == 1)) { true_pos = true_pos + 1;  }
	}
	true_neg = true_neg - sequence_number;	//binary array contains #seq extra columns that are 0/0
	total_size = true_neg + false_neg + false_pos + true_pos;

	if (true_pos == 0 && (false_pos == 0 || false_neg == 0)) {
		distance = 0;
	} else {
		nominator = (true_pos * true_neg) - (false_neg * false_pos);
		divider = ((true_pos + false_neg) * (true_neg + false_pos) * (true_pos + false_pos) * (true_neg + false_neg));
		distance = nominator / sqrt(divider);
		if (distance < 0) { distance = 0; }
	}

	return(distance);
};
//*********************************************************************************************************//
void BinaryPatternSequence::MergeArrays(BinaryPatternSequence * removable_array) {
        for (unsigned int i = 0; i < array_size ; ++i) {
                if((*removable_array).binary_patsequence[i] != 0) { 
			binary_patsequence[i] = (*removable_array).binary_patsequence[i]; 
		}
                if((*removable_array).starting_positions[i] != 0) {
			starting_positions[i] = (*removable_array).starting_positions[i];
		}
	}
};
//*********************************************************************************************************//
void BinaryPatternSequence::RenameElement(unsigned long counter, string &new_element, string &element1) {
	char buffer[50];
	string temp = "";

	sprintf(buffer, "%lu", counter);
	temp = buffer;

	element1 = original_pattern;
	original_pattern = "node_nro" + temp;
	new_element = original_pattern;
};
//*********************************************************************************************************//
void BinaryPatternSequence::InitializeMatrix( string comb_sequence, double at_prior_prob = 0.5, double cg_prior_prob = 0.5 ) {
	string temp_pattern = "";
	string mins_pattern = "";
	string mtrx_pattern = "";
	string nucleotide = "";

	long pat_length(0);
	Sequence pattern;
        
	for (unsigned int i = 0; i < array_size ; ++i) {
		pat_length = starting_positions[i];
		if (pat_length > 0) {
			temp_pattern = comb_sequence.substr(i, pat_length);
			mtrx_pattern = pattern.UpperCase(temp_pattern);
		}
		if (pat_length < 0) {
			temp_pattern = comb_sequence.substr(i, (-1 * pat_length));
			temp_pattern = pattern.UpperCase(temp_pattern);
			mtrx_pattern = pattern.RevComp(temp_pattern);
		}
		if (pat_length != 0) {
			for ( unsigned int j = 0; j < pattern_length; ++j) {
				if (mtrx_pattern.at(j) == 'A' || mtrx_pattern.at(j) == 'a') { 
					(*pat_matrix).SetElement(0, j, ((*pat_matrix).GetElement(0, j) + 1)); 
				} else if (mtrx_pattern.at(j) == 'C' || mtrx_pattern.at(j) == 'c') { 
					(*pat_matrix).SetElement(1, j, ((*pat_matrix).GetElement(1, j) + 1)); 
				} else if (mtrx_pattern.at(j) == 'G' || mtrx_pattern.at(j) == 'g') { 
					(*pat_matrix).SetElement(2, j, ((*pat_matrix).GetElement(2, j) + 1)); 
				} if (mtrx_pattern.at(j) == 'T' || mtrx_pattern.at(j) == 't') { 
					(*pat_matrix).SetElement(3, j, ((*pat_matrix).GetElement(3, j) + 1)); 
				}
			}
		}
	}
	(*pat_matrix).CalculatePatternMatrix( at_prior_prob, cg_prior_prob);
};
//*********************************************************************************************************//
std::string BinaryPatternSequence::PrintPatternMatrix( FILE *outstream ) {
	std::string consensus = "";
	fprintf (outstream, ">%s\n", original_pattern.c_str());
	consensus = (*pat_matrix).PrintPatternMatrix( outstream );
	return(consensus);
};
//*********************************************************************************************************//
void BinaryPatternSequence::MergePatternMatrix( BinaryPatternSequence * remove_element ) {
	(*pat_matrix).MergePatternMatrix( (*remove_element).pat_matrix );
};
//*********************************************************************************************************//
