#include <stdlib.h>
#include <math.h>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <string>
#include <vector>
#include "Sequence.h"
#include "BinaryPatternSequence.h"
#include "PatternMatrix.h"

using namespace std;

//*********************************************************************************************************//
PatternMatrix::PatternMatrix() {
	has_data = 1;
	row = 4;
};
//*********************************************************************************************************//
PatternMatrix::~PatternMatrix() {
	has_data = 0;
	if (pattern_matrix != NULL) { FreePatternMatrix(); }
};
//*********************************************************************************************************//
void PatternMatrix::FreePatternMatrix() {
	for( unsigned int i = 0; i < row; ++i) {
		delete [] pattern_matrix[i];
		pattern_matrix[i] = NULL;
	}
	delete [ ] pattern_matrix;
	pattern_matrix = NULL;
};
//*********************************************************************************************************//
void PatternMatrix::SetPatternMatrix(unsigned long column) {
	col = column;

	pattern_matrix = new double * [row];
	for( unsigned int i = 0; i < row; ++i) {
		pattern_matrix[i] = new double[col];
	}
	for( unsigned int i = 0; i < 4; ++i) {
		for( unsigned int j = 0; j < col; ++j) {
			pattern_matrix[i][j] = 0;
		}
	}
};
//*********************************************************************************************************//
void PatternMatrix::SetElement(unsigned long row_pos, unsigned long col_pos, double value) {
	if ((col_pos < col) && (row_pos < row)) {
		pattern_matrix[row_pos][col_pos] = value;
	}
};
//*********************************************************************************************************//
double PatternMatrix::GetElement(unsigned long row_pos, unsigned long col_pos ) {
	double value(0);
	if ((col_pos < col) && (row_pos < row)) {
		value = pattern_matrix[row_pos][col_pos];
	}
	return value;
};
//*********************************************************************************************************//
void PatternMatrix::MergePatternMatrix(PatternMatrix * removable_matrix ) {
	unsigned long matrix_size(0);
	unsigned long matrix_size_remove(0);
	unsigned long temp_matrix_size(0);
	unsigned long visited_number(0);
	unsigned long best_visited(0);
	unsigned long best_loop(0);
	long first_visit(-1);

	double column_max(0);
	double column_max_sum(0);
	double matrix_weight(0);
	double best_alignment(0);
	
	double * visited_columns;

	matrix_size = col;
	matrix_size_remove = (*removable_matrix).col;
	temp_matrix_size = (matrix_size + (2 * (matrix_size_remove -1)));

	PatternMatrix *temp_matrix;
	temp_matrix = new PatternMatrix;
	(*temp_matrix).SetPatternMatrix(temp_matrix_size);	

	PatternMatrix *best_matrix;
	best_matrix = new PatternMatrix;
	(*best_matrix).SetPatternMatrix(temp_matrix_size);

	visited_columns = new double[temp_matrix_size];

	for (unsigned int i = 0; i < 2; ++i) {
		for (unsigned int j = 0; j < (temp_matrix_size- (matrix_size_remove -1)); ++j) {
			for (unsigned int k = 0; k < temp_matrix_size; ++k) {
				visited_columns[k] = 0;
			}

			for( unsigned int k = 0; k < 4; ++k) {
				for( unsigned int l = 0; l < temp_matrix_size; ++l) {
					if ((i == 0) && (l >= (matrix_size_remove - 1)) && (l < (matrix_size + matrix_size_remove - 1))) {
						(*temp_matrix).SetElement(k, l, pattern_matrix[k][(l - matrix_size_remove + 1)]);
						visited_columns[l] = 1;
					} else if ((i == 1) && (l >= (matrix_size_remove - 1)) && (l < (matrix_size + matrix_size_remove - 1))) {
						(*temp_matrix).SetElement(k, l, pattern_matrix[3-k][(matrix_size -1 ) - (l - matrix_size_remove + 1)]);
						visited_columns[l] = 1;
					} else {
						(*temp_matrix).SetElement(k, l, 0);
					}
				}
			}
			for( unsigned int k = 0; k < 4; ++k) {
				for( unsigned int l = 0; l < (matrix_size_remove); ++l) {
					if ((j + l) < temp_matrix_size) {
			(*temp_matrix).SetElement(k, (j+l), ((((*temp_matrix).GetElement(k,(j+l)) + (*removable_matrix).GetElement(k,l))))/2);
			//(*temp_matrix).SetElement(k, (j+l), ((*temp_matrix).GetElement(k,(j+l)) + (*removable_matrix).GetElement(k,l)));
						visited_columns[(j + l)] = 1;
					}
				}
			}
			
			column_max_sum = 0;
			for (unsigned int l = 0; l < temp_matrix_size; ++l) {
				column_max = -1000;
				for( unsigned int k = 0; k < 4; ++k) {
					matrix_weight = (*temp_matrix).GetElement(k, l);
					if (matrix_weight >= column_max) { column_max = matrix_weight; }
					
				}
				column_max_sum = column_max_sum + column_max;
			}

			first_visit 	= -1;
			visited_number 	= 0;
			for (unsigned int m = 0; m < temp_matrix_size; ++m) {
				if ( visited_columns[m] == 1) { ++visited_number; }
				if ( visited_columns[m] == 1 && first_visit == -1) { first_visit = m; }
			}

			column_max_sum = column_max_sum / visited_number;			

			if (column_max_sum >= best_alignment) { 
				best_alignment 	= column_max_sum; 
				best_visited 	= visited_number;
				best_loop 	=  first_visit;
				for( unsigned int k = 0; k < 4; ++k) {
					for (unsigned int l = 0; l < temp_matrix_size; ++l) {
						(*best_matrix).SetElement(k, l, (*temp_matrix).GetElement(k, l));
					}
				}
			}
		}
	}

	FreePatternMatrix();
	SetPatternMatrix( best_visited );

	for( unsigned int k = 0; k < 4; ++k) {
		for (unsigned int l = 0; l < best_visited; ++l) {
			SetElement(k, l, (*best_matrix).GetElement(k, (l + best_loop)));
		}
	}

	(*best_matrix).FreePatternMatrix();
	(*temp_matrix).FreePatternMatrix();
	delete best_matrix;
	delete temp_matrix;
	delete [] visited_columns;
};
//*********************************************************************************************************//
void PatternMatrix::CalculatePatternMatrix( double at_prior_prob = 0.5, double cg_prior_prob = 0.5 ) {
	double matrix_value(0);
	double matrix_weight(0);
	double position_freq(0);
	double pattern_freq(0);
	double pseudo_weight(0);
	double prior_prob(0);
	
	for( unsigned int j = 0; j < col; ++j) {
		position_freq = (pattern_matrix[0][j] + pattern_matrix[1][j] + pattern_matrix[2][j] + pattern_matrix[3][j]);
		for( unsigned int i = 0; i < 4; ++i) {
			pattern_freq = pattern_matrix[i][j];
			pseudo_weight = 1;
                        
			if ((i == 0) || (i == 3)) { prior_prob = at_prior_prob / 2; }
			if ((i == 1) || (i == 2)) { prior_prob = cg_prior_prob / 2; }

			matrix_value = (pattern_freq + (pseudo_weight * prior_prob)) / (position_freq + pseudo_weight);
			matrix_weight = log(matrix_value / prior_prob);
			pattern_matrix[i][j] = matrix_weight;
            		
			//matrix_value = ((nuc_freq / total_freq) + (pseudo_weight * prior_prob)) / (1 + pseudo_weight);
		}
	}
};
//*********************************************************************************************************//
std::string PatternMatrix::PrintPatternMatrix( FILE *outstream  ) {
	std::vector <std::string> consensus_pattern;
	std::string consensus = "";
	std::string nucleotide = "";
	double matrix_weight(0);

	for( unsigned int i = 0; i < col; ++i) { consensus_pattern.push_back(""); }

	for( unsigned int j = 0; j < col; ++j) {
		for( unsigned int i = 0; i < row; ++i) {
			if (i == 0) { nucleotide = 'A'; }
			if (i == 1) { nucleotide = 'C'; }
			if (i == 2) { nucleotide = 'G'; }
			if (i == 3) { nucleotide = 'T'; }

			matrix_weight = pattern_matrix[i][j];
			if (matrix_weight > 0) { consensus_pattern[j] = consensus_pattern[j] + nucleotide; }
			fprintf (outstream, "%.2f\t", matrix_weight);
		}
		fprintf (outstream, "\n");
	}
	
	fprintf (outstream, "#consensus:\t");
	for (unsigned int i = 0; i < col; ++i){
		if (consensus_pattern[i] == "A" ) 		{ fprintf (outstream, "A"); consensus = consensus + "A"; }
		else if (consensus_pattern[i] == "C" ) 		{ fprintf (outstream, "C"); consensus = consensus + "C"; }
		else if (consensus_pattern[i] == "G" ) 		{ fprintf (outstream, "G"); consensus = consensus + "G"; }
		else if (consensus_pattern[i] == "T" ) 		{ fprintf (outstream, "T"); consensus = consensus + "T"; }
		else if (consensus_pattern[i] == "AC" ) 	{ fprintf (outstream, "M"); consensus = consensus + "M"; } 
		else if (consensus_pattern[i] == "AG" ) 	{ fprintf (outstream, "R"); consensus = consensus + "R"; }
		else if (consensus_pattern[i] == "AT" ) 	{ fprintf (outstream, "W"); consensus = consensus + "W"; }
		else if (consensus_pattern[i] == "CG" ) 	{ fprintf (outstream, "S"); consensus = consensus + "S"; }
		else if (consensus_pattern[i] == "CT" ) 	{ fprintf (outstream, "Y"); consensus = consensus + "Y"; }
		else if (consensus_pattern[i] == "GT" ) 	{ fprintf (outstream, "K"); consensus = consensus + "K"; }
		else if (consensus_pattern[i] == "ACG" ) 	{ fprintf (outstream, "V"); consensus = consensus + "V"; }
		else if (consensus_pattern[i] == "ACT" ) 	{ fprintf (outstream, "H"); consensus = consensus + "H"; }
		else if (consensus_pattern[i] == "AGT" ) 	{ fprintf (outstream, "D"); consensus = consensus + "D"; }
		else if (consensus_pattern[i] == "CGT" ) 	{ fprintf (outstream, "B"); consensus = consensus + "B"; }
		else 						{ fprintf (outstream, "N"); consensus = consensus + "N"; }
	}
	fprintf (outstream, "\n\n");
	return(consensus);
};
//*********************************************************************************************************//
