#include <stdio.h>
#include <iostream>
#include <string>
#include <stdlib.h>
#include <vector>
#include <cstdlib>
#include <fstream>
#include <unistd.h>
#include <getopt.h>

#include "Sequence.h"						//PROMOTER SEQUENCE CLASS 
#include "UserParameters.h"
#include "SequenceArray.h"
#include "BinaryPatternSequence.h"
#include "DistanceMatrix.h"
#include "ClusterResults.h"

namespace general {
	UserParameters param;                                   //PARAMETERS OBJECT
};

using namespace general;
using namespace std;
//*********************************************************************************************************//
unsigned long ReadSequenceInput(unsigned long&, string, std::vector <std::string>& seq_vec);
unsigned long ReadPatternInput(string, std::vector <std::string>& seq_vec);
unsigned long GetSmallestLength(unsigned long&, string);
void CalculateHiearchicalClustering(std::vector <std::string>&, std::vector <std::string>&, std::vector <ClusterResults>&, FILE*, FILE*);
void fnExit(void);
bool read_options(int, char);

int getopt(int argc, char *const argv[], const char *optstring);
int getopt_long(int argc, char *const argv[], const char *optstring, const struct option *longopts, int *longindex);
extern char *optarg;
extern int optind, opterr, optopt;

//*********************************************************************************************************//
void fnExit (void) {
	//string delete_file;
	//delete_file = param.output_file +".mtrx";        
	//remove(delete_file.c_str());
	//delete_file = param.output_file +".tree";
	//remove(delete_file.c_str());
        //delete_file = param.output_file +".pats";
        //remove(delete_file.c_str());
};
//*********************************************************************************************************//
unsigned long ReadPatternInput (string filename, vector<std::string>& seq_vec) {
	unsigned int counter(0);
	unsigned long pattern_length(0);
	string line = "";
	ifstream iFile (filename.c_str());

	if (iFile.is_open() && iFile.good()) {
		while ( (iFile.eof()) == 0 ) {
			getline(iFile, line);
			if ((line.size() > 0) && (line.at(0) != '>')) {
				{
					Sequence sequence;
					pattern_length = sequence.PromoterLength(line);
					if (pattern_length > 0) {
						seq_vec.push_back(line);
						++counter;
					}
				}
			}
		}
	} else {
		fprintf(stderr, "ERROR IN READING FILE: %s\n", filename.c_str());
	}
	iFile.close();
	counter = seq_vec.size();
	return ( counter );
}
//*********************************************************************************************************//
unsigned long ReadSequenceInput (unsigned long& smallest_promoter, string filename, vector<std::string>& seq_vec) {
	unsigned int counter(0);
	unsigned long promo_len(0);
	string line = "";                                       //INPUT PROMOTER SEQUENCE OR HEADER LINE
	string line_seq = "";
	Sequence sequence;
	
	ifstream iFile (filename.c_str());
	if (iFile.is_open() && iFile.good()) {
		while ( (iFile.eof()) == 0 ) {
			getline(iFile, line);
			if (line.size() > 0 && line.at(0) == '>') {
				if (counter > 0 && line_seq.size() > 3) { seq_vec.push_back(line_seq); }
				++counter;

				promo_len = GetSmallestLength( smallest_promoter, line_seq );
				line_seq = "";
			} else if ((line.size() > 0)  && (line.at(0) != '>') && (counter > 0)) {
				line_seq = line_seq + sequence.UpperCase(line);
			}
		}
		if ((counter > 0) && (line_seq.size() > 0)) {
			seq_vec.push_back(line_seq);
		}

		promo_len = GetSmallestLength( smallest_promoter, line_seq );
		line_seq = "";
	} else {
		fprintf(stderr, "ERROR IN READING FILE: %s\n", filename.c_str());
	}

	iFile.close();
	counter = seq_vec.size();
	return ( counter );
};
//*********************************************************************************************************//
unsigned long GetSmallestLength(unsigned long& smallest_promoter, string line_seq ) {
	unsigned long promo_len(0);

	Sequence sequence;
	promo_len = sequence.PromoterLength(line_seq);
	if ((promo_len < smallest_promoter) && (promo_len > 3)) {
		smallest_promoter = promo_len;
	}
	return ( promo_len );
};
//*********************************************************************************************************//
bool read_options(int argc, char ** argv) {
	bool options_ok(1);
	bool bool1(0), bool2(0), bool3(0), bool4(0), bool0(0);
	int co;

	struct option longopts[] = {
   		{ "seq_file1",	1,	NULL,	'A' },
		{ "seq_file2",  1,      NULL,   'B' },
		{ "pat_file",   1,      NULL,   'M' },
   		{ "ofile",	1,	NULL,	'O' },
   		{ "help",	0,	NULL,	'h' },		
   		{ 0, 0, 0, 0 }
	};

	while ((co = getopt_long(argc, argv, ":A:B:O:L:M:hW;", longopts, NULL)) != -1) {
		switch (co) {
		case 'A': 
			param.sequence_file1 		= optarg;
			bool0				= 1;
			param.has_two_files		= 0;
			break;
		case 'B':
			param.sequence_file2		= optarg;
			param.has_two_files		= 1;
			bool1				= 1;
			break;
		case 'O': 
			param.output_file 		= optarg; 	
			param.output_matrix		= param.output_file + ".mtrx";
			param.output_hctree		= param.output_file + ".tree";
			param.output_pattern		= param.output_file + ".pats";
			bool2 				= 1; 	
			break;
		case 'L': 
			param.max_pos 			= atoi(optarg); 
			bool3 				= 1; 	
			break;
		case 'M': 
			param.pattern_file 		= optarg;  	
			bool4				= 1;
			break;
		case 'h': 
			options_ok 			= 0; 					
			break;
		case 0: 							
			break;

		case ':':
			fprintf(stderr, "%s: OPTION `-%c' REQUIRES AN ARGUMENT\n", argv[0], optopt); 
			options_ok = 0; 
			break;
		case '?':
			default:
			fprintf(stderr, "%s: OPTION `-%c' IS INVALID: IGNORED\n",argv[0], optopt); 
			options_ok = 0; 
			break;
		}
	}

	if (options_ok == 1) {
		if (bool0 == 0 || bool4 == 0) {
			fprintf(stderr, "ERROR: NO INPUT FILES (TYPE -A -M OR --seq_file --pat_file)\n"); 
			options_ok = 0;
		}
		if (bool2 == 0 || bool3 == 0) {
			fprintf(stderr, "ERROR: OBLIGATORY PARAMETERS MISSING\n"); 
			options_ok = 0;
		}
	}

	return options_ok;
};
//*********************************************************************************************************//
void CalculateHiearchicalClustering(vector<std::string>& seq_vec,  vector<std::string>& pat_vec, vector<ClusterResults>& pat_res, FILE *outstream_mtrx, FILE *outstream_pats) {
	unsigned long pattern_counter1(0);
	unsigned long pattern_counter2(0);
	unsigned long pattern_total(0);
	unsigned long row_max_dist(0);
	unsigned long col_max_dist(0);
	unsigned long row_pat(0);
	unsigned long col_pat(0);
	unsigned long n_seqs = param.sequence_set_size;
	unsigned long n_pats = param.pattern_set_size;
	unsigned long counter(0);

	double pattern_dist(0);
	double combined_pair_value(0);
	double sequence_prob_at;
	double sequence_prob_cg;

	std::string combined_sequences;
	std::string new_element;
	std::string element1;
	std::string element2;
	std::string consensus;

	BinaryPatternSequence * binary_seq      	= NULL;
	BinaryPatternSequence * binary_first    	= NULL;
	BinaryPatternSequence * binary_last     	= NULL;
	BinaryPatternSequence * bin_iter1       	= NULL;
	BinaryPatternSequence * bin_iter2       	= NULL;
 
	{
		SequenceArray sequence_array;
		sequence_array.SetSequence(seq_vec, param.max_pos);
		combined_sequences = sequence_array.GetSequence();
		sequence_array.CalculateATGC(sequence_prob_at, sequence_prob_cg);
	}
	
	{
		binary_seq      = new BinaryPatternSequence [n_pats];
		binary_first    = &binary_seq[0];
		binary_last     = &binary_seq[n_pats];

		pattern_counter1 = 0;
		for (bin_iter1 = binary_first; bin_iter1 != binary_last; bin_iter1++) {
			(*(bin_iter1)).SetSize(combined_sequences.size());
			(*(bin_iter1)).SetArray((pattern_counter1 + 1), pat_vec[pattern_counter1], combined_sequences);
			(*(bin_iter1)).InitializeMatrix(combined_sequences, sequence_prob_at, sequence_prob_cg);
			(*(bin_iter1)).PrintPatternMatrix(outstream_pats);
			++pattern_counter1;
		}

		pattern_total = pattern_counter1;
		DistanceMatrix dist_mat;
		dist_mat.SetSize(n_pats, n_pats);

		pattern_counter1 = 0;
		pattern_counter2 = 0;
		for (bin_iter1 = binary_first; bin_iter1 != (binary_last -1); ++bin_iter1) {
			pattern_counter2 = pattern_counter1 + 1;
				for (bin_iter2 = (bin_iter1 + 1); bin_iter2 != binary_last; ++bin_iter2) {
				if (((*(bin_iter1)).has_data == 1) && ((*(bin_iter2)).has_data == 1)) {
					pattern_dist = (*(bin_iter1)).CalculateDistance(bin_iter2, n_seqs);
					dist_mat.SetElement(pattern_counter1, pattern_counter2, pattern_dist);
					dist_mat.SetElement(pattern_counter2, pattern_counter1, pattern_dist);
					++pattern_counter2;
				}
			}
			++pattern_counter1;
		}
		dist_mat.PrintMatrix(outstream_mtrx);
		
		for (unsigned int i = 0; i < (n_pats -1); ++i) {
			combined_pair_value = dist_mat.FindMax(row_max_dist, col_max_dist);
			dist_mat.RemoveElement(col_max_dist);

			counter = 0; row_pat = 0;
			for (unsigned int j = 0; j < n_pats; ++j) {
				if (((*(binary_first + j)).has_data == 1)) { ++counter; }
				if (counter == row_max_dist) { row_pat = (j + 1); }
			}

			counter = 0; col_pat = 0;
			for (unsigned int j = 0; j < n_pats; ++j) {	
				if (((*(binary_first + j)).has_data == 1)) { ++counter; }
				if (counter == col_max_dist) { col_pat = (j + 1); }
			}

			(*(binary_first + row_pat)).MergeArrays( binary_first + col_pat );
			(*(binary_first + row_pat)).RenameElement((pattern_total + long(i) + 1), new_element, element1);

			(*(binary_first + row_pat)).MergePatternMatrix( binary_first + col_pat );
			consensus = (*(binary_first + row_pat)).PrintPatternMatrix( outstream_pats );				
			
			element2 = (*(binary_first + col_pat)).original_pattern;
			(*(binary_first + col_pat)).FreeMemory();

			pattern_counter1 = 0;
			for (bin_iter1 = binary_first; bin_iter1 != (binary_last -1); ++bin_iter1) {
				if (((*(bin_iter1)).has_data == 1) && ((*(binary_first + row_pat)).has_data == 1)) {
					if((binary_first + row_pat) != bin_iter1) {
						pattern_dist = (*(bin_iter1)).CalculateDistance((binary_first + row_pat), n_seqs);
						dist_mat.SetElement(row_max_dist, pattern_counter1, pattern_dist);
						dist_mat.SetElement(pattern_counter1, row_max_dist, pattern_dist);
					}
					++pattern_counter1;
				}
			}

			{
				ClusterResults this_result;
				this_result.SetElement(new_element, element1, element2, consensus, combined_pair_value);
				pat_res.push_back(this_result);
			}

			dist_mat.PrintMatrix(outstream_mtrx);
		}
		delete [ ] binary_first;
		dist_mat.FreeMemory();
	} 
};
//*********************************************************************************************************//
int main (int argc, char** argv) {
	unsigned long smallest_promoter(10000000);
	unsigned long promoter_number(0);
	unsigned long pattern_number(0);

	double distace_value(0);
	
	std::string comb_element 	= "";
	std::string element1 		= "";
	std::string element2 		= "";
	std::string consensus 		= "";

	std::vector <std::string> sequence_set;
	std::vector <std::string> pattern_set;
	std::vector <ClusterResults> result_vector;

        srand((unsigned)time(NULL));
	atexit(fnExit);
	bool options_ok;				//Bool to check inputs
	options_ok = read_options(argc, argv);		//read input options

	if (options_ok == 0) { 
		fprintf (stderr, "\nFOLLOWING PARAMETERS ARE OBLIGATORY\n");
		fprintf (stderr, "-L <number>\tpromoter length (bp) (default is 0 = length of the shortest promoter)\n");
		fprintf (stderr, "-M or --seq_file <string>\tPatterns to be clustered\n");
		fprintf (stderr, "-A or --pat_file1 <string>\tthe first input file\n");
		fprintf (stderr, "-B or --pat_file2 <string>\tthe second input file\n");
		fprintf (stderr, "-O or --ofile <string>\tprefix of the output files (.tree .pats .mtrx)\n");

                if (!(param.sequence_file1.empty())) {
			FILE *inputstrm_file1;
			inputstrm_file1 = fopen(param.sequence_file1.c_str(), "w+");
			fprintf(inputstrm_file1, "No fasta formatted files in input file 1");
			fclose(inputstrm_file1);
		}

		if (!(param.sequence_file2.empty())) {
			FILE *inputstrm_file2;
			inputstrm_file2 = fopen(param.sequence_file2.c_str(), "w+");
			fprintf(inputstrm_file2, "No fasta formatted files in input file 2");
			fclose(inputstrm_file2);
		}
		
		exit(0);
	} else {
		promoter_number = ReadSequenceInput( smallest_promoter, param.sequence_file1, sequence_set );
		if (param.has_two_files == 1) { 
			promoter_number = ReadSequenceInput( smallest_promoter, param.sequence_file2, sequence_set ); 
		}
		
		if (promoter_number > 0) {
			param.sequence_set_size	= promoter_number ;
			if (param.max_pos == 0) { param.max_pos = smallest_promoter; }
		}
		pattern_number = ReadPatternInput( param.pattern_file, pattern_set );
		if (pattern_number > 1) {
			param.pattern_set_size	= pattern_number;
		}
		if (param.sequence_set_size == 0 || param.pattern_set_size == 0) {
			fprintf(stderr, "ERROR: NO INPUT FILES (TYPE -A -M OR --seq_file --pat_file)\n");
			exit(0);
		} else {
			FILE *outstream_tree;
			outstream_tree = fopen(param.output_hctree.c_str(), "w");
			if (outstream_tree == NULL) { fprintf (stderr, "ERROR 1: FILE NOT FOUND %s\n", param.output_hctree.c_str()); exit(0); }
			if (ferror(outstream_tree)) { fprintf (stderr, "ERROR 2: FILE NOT FOUND %s\n", param.output_hctree.c_str()); exit(0); }
        
			FILE *outstream_mtrx;
			outstream_mtrx = fopen(param.output_matrix.c_str(), "w");
			if (outstream_mtrx == NULL) { fprintf (stderr, "ERROR 1: FILE NOT FOUND %s\n", param.output_matrix.c_str()); exit(0); }
			if (ferror(outstream_mtrx)) { fprintf (stderr, "ERROR 2: FILE NOT FOUND %s\n", param.output_matrix.c_str()); exit(0); }
			
			FILE *outstream_pats;
			outstream_pats = fopen(param.output_pattern.c_str(), "w");
			if (outstream_pats == NULL) { fprintf (stderr, "ERROR 1: FILE NOT FOUND %s\n", param.output_pattern.c_str()); exit(0); }
			if (ferror(outstream_pats)) { fprintf (stderr, "ERROR 2: FILE NOT FOUND %s\n", param.output_pattern.c_str()); exit(0); }

			CalculateHiearchicalClustering(sequence_set, pattern_set, result_vector, outstream_mtrx, outstream_pats);
			for(unsigned int i = 0; i < result_vector.size(); ++i) {
				result_vector[i].GetElement(comb_element, element1, element2, consensus, distace_value);		
				fprintf(outstream_tree, "%s\t%s\t%s\t%.3f\t%s\n", comb_element.c_str(), element1.c_str(), element2.c_str(), distace_value, consensus.c_str());
			}
			fclose(outstream_tree);
			fclose(outstream_mtrx);
		}
	}

	return 1;
};
//*********************************************************************************************************//

